!	Node RG -- Node Report Generator
!	Copyright (C) 2000,2003 Nils M Holm. All rights reserved.
!	See the file LICENSE for conditions of use.
!
!	2000-10-24	Started implementation
!	2000-10-28	Internal Beta Release (Hanne)
!	2000-10-29	Fields were never cleared in newsheet()
!	2000-10-29	fld[F_LEN] overrode fld[F_FLGS] & FF_NOFILL
!	2001-12-05	Finished implementation of zoom mode

module node_rptgen(t3x, string, iostream, system, ttyctl, memory, config, menu,
		prompt, form);

object	t[t3x], str[string], sys[system], tty[ttyctl], mem[memory], mnu[menu],
	prm[prompt], frm[form];

const	RECORDSIZE	= 1024;
const	PATHLEN		= 256;
const	NAMELEN		= 30;
const	POOLSIZE	= 8192;
const	MAXLINES	= 60;
const	MAXFIELDS	= 16;
const	BUFSIZE		= 1024;
const	MAXSHEETS	= 16;
const	TAGTBLSIZE	= 513;
const	NDXLEN		= 16;
const	I_KEY		= 15;

const	NLINES		= 25,
	NCOLS		= 80,
	NCXLINES	= 18,
	NCXCOLS		= 64;

object	Cnf[config];

struct	FIELD = F_X, F_Y, F_TYPE, F_NAME, F_LEN, F_FLGS, F_NEXT;
struct	TYPES = T_LIT, T_REF;
const	FF_OFF1 = 0x0001,
	FF_OFF2 = 0x0002,
	FF_OFF3 = 0x0004,
	FF_FILL = 0x0008,
	FF_TRIM = 0x0010;

object	Rpt[iostream];
var	Rptbuf::BUFSIZE;
var	Rptfile::PATHLEN+1;
var	Pool::POOLSIZE;
var	Dbfd, Ndxfd, Tagfd;
var	Record::RECORDSIZE, Dataset[MAXFIELDS];
var	Descriptor::RECORDSIZE;
var	Confpath::PATHLEN+1;
var	Basename::PATHLEN+1, Dbpath::PATHLEN+1;
var	Xmax, Ymax, Mtop, Topline;
var	Blanks::NCOLS+2, Hline::NCOLS+2;
var	Sheet[MAXLINES];
var	Sheetname::NAMELEN+1;
var	Sheetpath::PATHLEN+1;
var	Sheetindex[MAXSHEETS+1], Shndxbuf::NAMELEN+1*MAXSHEETS, Nsheets;
var	Sheetnlist[MAXSHEETS+1], Snlbuf[MAXSHEETS*2], Snt;
var	Top, Xpos, Ypos, Offset;
var	Indent, Width, Spacing, Pagesize, Absmode;
var	Nfields;
var	Row;
var	Taglist[MAXFIELDS];
var	Fieldnames[MAXFIELDS+1], Fmbuffer[MAXFIELDS*2];
var	Floating;
var	Posted;
var	Modified;
var	Tagtbl::TAGTBLSIZE;

var	O_hints,
	O_warnings,
	O_zoom,
	O_dosflag,
	O_rg_flags,
	O_rg_addcr;
var	O_borderc::NAMELEN+1;
var	O_rgpath::PATHLEN+1,
	O_rg_print::PATHLEN+1,
	O_rg_view::PATHLEN+1;
var	C_background,
	C_alert,
	C_menuframe,
	C_menutext,
	C_menubar,
	C_menuhigh,
	C_infoframe,
	C_infotext,
	C_promptframe,
	C_prompttext,
	C_promptbar,
	C_editframe,
	C_editprompt,
	C_edittext,
	C_editbar,
	C_browseframe,
	C_browsetext,
	C_browsehead,
	C_browsebar,
	C_browsetag,
	C_logo	,
	C_titlehi,
	C_titlelo,
	C_helpframe,
	C_helptext,
	C_helpkey,
	C_tagind,
	C_marker,
	C_limitind;

configure() do
	const   TXT = config.T_STRING,
		NUM = config.T_NUMERIC,
		COL = config.T_COLOR;
	const	C00 = ttyctl.F_GREY|ttyctl.F_BRIGHT|ttyctl.B_BLUE,
		C01 = ttyctl.F_YELLOW|ttyctl.F_BRIGHT|ttyctl.B_RED,
		C02 = ttyctl.B_CYAN,
		C03 = ttyctl.F_GREY|ttyctl.F_BRIGHT|ttyctl.B_CYAN,
		C04 = ttyctl.F_GREY|ttyctl.F_BRIGHT|ttyctl.B_YELLOW,
		C05 = ttyctl.F_CYAN|ttyctl.F_BRIGHT|ttyctl.B_CYAN,
		C06 = ttyctl.B_YELLOW,
		C07 = ttyctl.F_GREY|ttyctl.F_BRIGHT|ttyctl.B_YELLOW,
		C08 = ttyctl.B_CYAN,
		C09 = ttyctl.F_GREY|ttyctl.F_BRIGHT|ttyctl.B_CYAN,
		C10 = ttyctl.F_GREY|ttyctl.F_BRIGHT|ttyctl.B_YELLOW,
		C11 = ttyctl.B_GREY,
		C12 = ttyctl.F_RED|ttyctl.B_GREY,
		C13 = ttyctl.F_BLACK|ttyctl.B_GREY,
		C14 = ttyctl.F_GREY|ttyctl.F_BRIGHT|ttyctl.B_YELLOW,
		C15 = ttyctl.B_GREY,
		C16 = ttyctl.F_BLACK|ttyctl.F_BRIGHT|ttyctl.B_GREY,
		C17 = ttyctl.F_RED|ttyctl.B_GREY,
		C18 = ttyctl.F_GREY|ttyctl.F_BRIGHT|ttyctl.B_YELLOW,
		C19 = ttyctl.F_CYAN|ttyctl.F_BRIGHT|ttyctl.B_GREY,
		C20 = ttyctl.B_CYAN,
		C21 = ttyctl.F_CYAN|ttyctl.F_BRIGHT|ttyctl.B_BLUE,
		C22 = ttyctl.F_CYAN|ttyctl.B_BLUE,
		C23 = ttyctl.B_CYAN,
		C24 = ttyctl.F_BLUE|ttyctl.B_CYAN,
		C25 = ttyctl.F_RED|ttyctl.B_CYAN,
		C26 = ttyctl.F_BRIGHT|ttyctl.F_CYAN|ttyctl.B_YELLOW,
		C27 = ttyctl.F_BRIGHT|ttyctl.F_GREY|ttyctl.B_GREY,
		C28 = ttyctl.F_BRIGHT|ttyctl.F_CYAN|ttyctl.B_GREY;

	Cnf.init([
[ "color-background",	COL, C00,		@C_background	],
[ "color-alert",	COL, C01,		@C_alert	],
[ "color-menuframe",	COL, C02,		@C_menuframe	],
[ "color-menutext",	COL, C03,		@C_menutext	],
[ "color-menubar",	COL, C04,		@C_menubar	],
[ "color-menuhigh",	COL, C05,		@C_menuhigh	],
[ "color-infoframe",	COL, C06,		@C_infoframe	],
[ "color-infotext",	COL, C07,		@C_infotext	],
[ "color-promptframe",	COL, C08,		@C_promptframe	],
[ "color-prompttext",	COL, C09,		@C_prompttext	],
[ "color-promptbar",	COL, C10,		@C_promptbar	],
[ "color-editframe",	COL, C11,		@C_editframe	],
[ "color-editprompt",	COL, C12,		@C_editprompt	],
[ "color-edittext",	COL, C13,		@C_edittext	],
[ "color-editbar",	COL, C14,		@C_editbar	],
[ "color-browseframe",	COL, C15,		@C_browseframe	],
[ "color-browsetext",	COL, C16,		@C_browsetext	],
[ "color-browsehead",	COL, C17,		@C_browsehead	],
[ "color-browsebar",	COL, C18,		@C_browsebar	],
[ "color-browsetag",	COL, C19,		@C_browsetag	],
[ "color-logo",		COL, C20,		@C_logo		],
[ "color-titlehi",	COL, C21,		@C_titlehi	],
[ "color-titlelo",	COL, C22,		@C_titlelo	],
[ "color-helpframe",	COL, C23,		@C_helpframe	],
[ "color-helptext",	COL, C24,		@C_helptext	],
[ "color-helpkey",	COL, C25,		@C_helpkey	],
[ "color-tagind",	COL, C26,		@C_tagind	],
[ "color-marker",	COL, C27,		@C_marker	],
[ "color-limitind",	COL, C28,		@C_limitind	],
[ "border-chars",	TXT, "\s\s\s\s\s\s",	@O_borderc	],
[ "show-hints",		NUM, 1,			@O_hints	],
[ "show-warnings",	NUM, 1,			@O_warnings	],
[ "zoom-mode",		NUM, 0,			@O_zoom		],
[ "report-generator",	TXT, "./node_rg",	@O_rgpath	],
[ "rg-print-command",	TXT, "/usr/bin/lp",	@O_rg_print	],
[ "rg-view-command",	TXT, "/usr/bin/more",	@O_rg_view	],
[ "rg-add-cr",		NUM, 1,			@O_rg_addcr	],
[ "rg-default-flags",	NUM, FF_FILL|FF_TRIM,	@O_rg_flags	],
	0 ]);
end

clear() do
	tty.color(C_background);
	tty.clear();
end

setup() do var i;
	ie (O_zoom) do
		Xmax := NCXCOLS;
		Ymax := NCXLINES;
	end
	else do
		Xmax := NCOLS;
		Ymax := NLINES;
	end
	Mtop := Ymax-4;
	Topline := 2;
	for (i=0,NCOLS) Blanks::i := '\s';
	for (i=0,NCOLS) Hline::i := O_borderc::4;
	Blanks::i := 0;
	Blanks::(i+1) := 0;
	clear();
end

init() do var buf::80;
	configure();
	! If COMSPEC is defined, we're probably running under DOS
	O_dosflag := t.getenv("COMSPEC", buf, 80) > 0;
	Dbfd := %1;
	Ndxfd := %1;
	Record::RECORDSIZE := 0;
	Confpath::0 := 0;
	Floating := 0;
	Posted := 0;
	str.copy(Sheetname, "*default*");
	Nsheets := 0;
	sys.init();
	tty.init();
	tty.mode(1);
end

fini() do
	if (Dbfd \= %1) t.close(Dbfd);
	if (Ndxfd \= %1) t.close(Ndxfd);
	tty.color(ttyctl.F_GREY|ttyctl.B_BLACK);
	tty.clear();
	tty.mode(0);
	tty.fini();
	sys.fini();
end

_hintbar(msg) do
	tty.color(C_infotext);
	tty.move(Xmax/2-str.length(msg)/2-1, Ymax-2);
	tty.writec('\s');
	tty.writes(msg);
	tty.writec('\s');
end

hintbar(msg) do
	if (\O_hints) return;
	tty.move(0, Ymax-2);
	tty.color(C_background);
	tty.clreol();
	_hintbar(msg);
end

message(msg) do
	tty.move(0, Ymax-4);
	tty.color(C_background);
	tty.clreol();
	Posted := msg;
	if (\msg) return;
	tty.color(C_alert);
	tty.move(Xmax/2-str.length(msg)/2-4, Ymax-4);
	tty.writes(">>> ");
	tty.writes(msg);
	tty.writes(" <<<");
end

initsheet() do var i;
	mem.init(Pool, POOLSIZE);
	for (i=0, MAXLINES) Sheet[i] := 0;
	Xpos := 0;
	Ypos := 0;
	Top := 0;
	Offset := 0;
	Indent := 4;
	Width := 72;
	Spacing := 0;
	Absmode := 0;
	Pagesize := 60;
	Modified := 0;
end

drawfield(fld) do var buffer::NCOLS+1, k;
	if (fld[F_X] < Offset \/ fld[F_X] > Xmax-Indent) return;
	tty.move(fld[F_X]+Indent-Offset, fld[F_Y]-Top+Topline);
	ie (fld[F_TYPE] = T_LIT) do
		tty.color(C_browsetext);
		ie (fld[F_NAME])
			str.copy(buffer, fld[F_NAME]);
		else
			buffer::0 := 0;
		k := str.length(buffer);
		if (k > fld[F_LEN]) do
			k := fld[F_LEN];
			buffer::k := 0;
		end
		if (k + fld[F_X] + Indent - Offset > Xmax) do
			k := Xmax-fld[F_X]-Indent + Offset;
			buffer::k := 0;
			k := fld[F_LEN];
		end
		tty.writes(buffer);
	end
	else do
		tty.color(C_browsebar);
		ie (fld[F_NAME])
			str.copy(buffer, fld[F_NAME]);
		else
			buffer::0 := 0;
		k := str.length(buffer);
		if (k > fld[F_LEN]) do
			k := fld[F_LEN];
			buffer::k := 0;
		end
		if (k + fld[F_X] + Indent - Offset > Xmax) do
			k := Xmax-fld[F_X]-Indent + Offset;
			buffer::k := 0;
			k := fld[F_LEN];
		end
		tty.writes(buffer);
		tty.writes(@Blanks::(NCOLS-fld[F_LEN]+k));
		tty.move(fld[F_X]+Indent, fld[F_Y]-Top+Topline);
	end
end

drawline(y, fld) do
	tty.move(Indent, y+Topline);
	tty.color(C_browsetext);
	tty.writes(Width + Indent < Xmax->
		@Blanks::(NCOLS-Width):
		@Blanks::(Ncols-Xmax+Indent));
	while (fld) do
		drawfield(fld);
		fld := fld[F_NEXT];
	end
end

drawlines(first, limit) do var i;
	for (i=first, limit)
		drawline(i, Sheet[i+Top]);
end

drawsheet() do
	tty.move(Indent+4, Topline-1);
	tty.color(C_background);
	tty.clreol();
	tty.color(C_browsehead);
	tty.writec('\s');
	tty.writes(Sheetname);
	tty.writec('\s');
	drawlines(0, Mtop-3);
end

linkfield(y, fld) do
	fld[F_NEXT] := Sheet[y];
	Sheet[y] := fld;
	Modified := 1;
end

newfield(x, y, lnk) do
	var	new;

	new := mem.alloc(FIELD*t.bpw());
	if (\new) do
		message("Not enough memory for new field");
		return 0;
	end
	new[F_X] := x;
	new[F_Y] := y;
	new[F_TYPE] := T_LIT;
	new[F_NAME] := 0;
	new[F_LEN] := 0;
	new[F_FLGS] := O_rg_flags;
	if (lnk) linkfield(y, new);
	return new;
end

unlinkfield(fld) do var p, pp;
	pp := @Sheet[fld[F_Y]];
	p := pp[0];
	while (p /\ p \= fld) do
		pp := @p[F_NEXT];
		p := p[F_NEXT];
	end
	if (\p) do
		message("OOPS! Failed to unlink field!");
		return 0;
	end
	pp[0] := p[F_NEXT];
	Modified := 1;
	return p;
end

deletefield(fld, ulnk) do
	if (ulnk) do
		fld := unlinkfield(fld);
		if (\fld) return;
	end
	if (fld[F_TYPE] = T_LIT /\ fld[F_NAME]) mem.free(fld[F_NAME]);
	mem.free(fld);
	Modified := 1;
end

savestr(s) do var new, k;
	k := str.length(s)+1;
	new := mem.alloc(k);
	if (\new) do
		message("Not enough memory to save string");
		return 0;
	end
	t.memcopy(new, s, k);
	return new;
end

findref(s) do var i;
	for (i=0, Nfields)
		if (\str.comp(Taglist[i], s))
			return Taglist[i];
	return 0;
end

copyfield(fld) do var new;
	new := newfield(fld[F_X], fld[F_Y], 0);
	if (\new) return 0;
	new[F_TYPE] := fld[F_TYPE];
	new[F_NAME] := fld[F_NAME];
	new[F_LEN] := fld[F_LEN];
	new[F_FLGS] := fld[F_FLGS];
	ie (fld[F_TYPE] = T_REF \/ \fld[F_NAME])
		new[F_NAME] := fld[F_NAME];
	else
		new[F_NAME] := savestr(fld[F_NAME]);
	if (Xpos + new[F_LEN] > Width)
		Xpos := Width - fld[F_LEN];
	Modified := 1;
	return new;
end

chg_field_type(fld) do var x, y, d, r, i;
	x := fld[F_X];
	y := fld[F_Y]-Top;
	if (x < 0) x := 0;
	if (x + 23 >= Width) x := Width-23;
	if (x + 23 > Xmax) x := Xmax-27;
	if (y < 0) y := 0;
	if (y >= Mtop-6) y := Mtop-7;
	d := fld[F_TYPE] = T_LIT-> 0: 1;
	hintbar("Select field type");
	r := mnu.select(x+Indent, y+Topline, d,
	[ (C_menutext), (C_menutext), (C_menutext), (C_menutext),
	  (C_menubar) ],
	O_borderc, [
		[ 'l', "Text literal    [L]" ],
		[ 'f', "Database field  [F]" ],
	0]);
	drawlines(y, y+4);
	if (r < 0) return 0;
	fld[F_TYPE] := r=0-> T_LIT: T_REF;
	return 1;
end

showframe() do var i;
	clear();
	if (C_background = C_browsetext) for (i=Topline, Mtop-1) do
		if (Indent) do
			tty.move(Indent-1, i);
			tty.writec('|');
		end
		if (Width+Indent < Xmax-1) do
			tty.move(Width+Indent, i);
			tty.writec('|');
		end
	end
end

chg_field_ref(fld) do var r;
	for (r=0, Nfields)
		if (fld[F_NAME] = Taglist[r]) leave;
	r := r >= Nfields-> 0: r;
	hintbar("Select field name");
	r := mnu.select(Indent, Topline, r,
	[ (C_menutext), (C_menutext), (C_menutext), (C_menutext),
	  (C_menubar) ],
	O_borderc, Fieldnames);
	showframe();
	drawsheet();
	if (r < 0) return;
	fld[F_NAME] := Taglist[r];
	return 1;
end

findlim(fld) do var lim, c0, cl;
	lim := Width;
	c0 := fld[F_X];
	fld := Sheet[fld[F_Y]];
	while (fld) do
		cl := fld[F_X];
		if (cl > c0 /\ cl < lim) lim := cl;
		fld := fld[F_NEXT];
	end
	return lim;
end

chg_field_name(fld) do var buffer::NCOLS+1, max, k;
	if (fld[F_TYPE] = T_REF)
		return chg_field_ref(fld);
	hintbar("Enter literal text to place in report");
	max := findlim(fld)-fld[F_X];
	ie (fld[F_NAME])
		str.copy(buffer, fld[F_NAME]);
	else
		buffer::0 := 0;
	k := prm.input(fld[F_X]+Indent, fld[F_Y]+Topline-Top, max, max+1,
		[ (C_menutext), (C_menutext), (C_menutext), (C_menutext),
		  (C_menubar) ],
		buffer, [ 2, '\r', ttyctl.K_ESC ]);
	drawline(fld[F_Y]-Top, fld);
	if (k = ttyctl.K_ESC) return 0;
	fld[F_NAME] := savestr(buffer);
	fld[F_LEN] := str.length(buffer);
	return 1;
end

showlen(k) do var buf::NCOLS+1;
	str.format(buf, "L:%D", [(k)]);
	tty.move(Indent, Mtop);
	tty.color(C_background);
	tty.clreol();
	tty.writes(buf);
end

chg_field_length(fld) do var k, len, x, y, w;
	hintbar("Use <- and -> to specify the field length");
	len := 0;
	len := fld[F_LEN];
	x := fld[F_X];
	y := fld[F_Y];
	if (x + fld[F_LEN] > Xmax-Indent) do
		Offset := Width-Xmax+Indent;
		drawsheet();
	end
	if (x < Offset) do
		Offset := 0;
		drawsheet();
	end
	tty.move(x+Indent-Offset, y-Top+Topline);
	tty.color(C_menutext);
	tty.writes(@Blanks::(Ncols-len));
	w := findlim(fld);
	while (1) do
		showlen(len);
		tty.move(x+len+Indent-Offset, y-Top+Topline);
		k := tty.readc();
		ie (k = ttyctl.K_LEFT \/ k = 'B'-'@') do
			if (len) len := len-1;
			tty.color(x+len = Width-1-> C_background:
				C_browseframe);
			tty.writec('\s');
		end
		else ie (k = ttyctl.K_RIGHT \/ k = 'F'-'@') do
			if (x+len < w) len := len+1;
			tty.color(C_menutext);
			tty.writec('\s');
		end
		else ie (k = ttyctl.K_ESC \/ k = ':') do
			return 0;
		end
		else if (k = '\r') do
			leave;
		end
	end
	fld[F_LEN] := len;
	return k \= ttyctl.K_ESC;
end

chg_field_flags(fld) do var x, y, d, r, i;
	x := fld[F_X];
	y := fld[F_Y]-Top;
	if (x < 0) x := 0;
	if (x + 23 >= Width) x := Width-23;
	if (x + 23 > Xmax) x := Xmax-27;
	if (y < 0) y := 0;
	if (y >= Mtop-8) y := Mtop-9;
	d := fld[F_TYPE] = T_LIT-> 0: 1;
	d := 3;
	while (1) do
		hintbar("Set field flags");
		r := mnu.select(x+Indent, y+Topline, d,
		[ (C_menutext), (C_menutext), (C_menutext), (C_menutext),
	  	(C_menubar) ],
		O_borderc, [
			[ 'o', (fld[F_FLGS] & FF_OFF3->
				"Offset ...... [+3] ":
		 	fld[F_FLGS] & FF_OFF2->
				"Offset ...... [+2] ":
		 	fld[F_FLGS] & FF_OFF1->
				"Offset ...... [+1] ":
				"Offset ...... [+0] ")	],
			[ 'f', (fld[F_FLGS] & FF_FILL->
				"Fill ........ [ON]":
				"Fill ........ [OFF]")	],
			[ 't', (fld[F_FLGS] & FF_TRIM->
				"Trim ........ [ON] ":
				"Trim ........ [OFF]")	],
			[ '\r', "      [DONE]       "	],
		0]);
		if (r < 0) leave;
		d := r;
		ie (r = 0) do
			ie (fld[F_FLGS] & FF_OFF1)
				fld[F_FLGS] := fld[F_FLGS] &
					~FF_OFF1 | FF_OFF2;
			else ie (fld[F_FLGS] & FF_OFF2)
				fld[F_FLGS] := fld[F_FLGS] &
					~FF_OFF2 | FF_OFF3;
			else ie (fld[F_FLGS] & FF_OFF3)
				fld[F_FLGS] := fld[F_FLGS] & ~FF_OFF3;
			else
				fld[F_FLGS] :=  fld[F_FLGS] | FF_OFF1;
		end
		else ie (r = 1) do
			ie (fld[F_FLGS] & FF_FILL)
				fld[F_FLGS] := fld[F_FLGS] & ~FF_FILL;
			else
				fld[F_FLGS] := fld[F_FLGS] | FF_FILL;
		end
		else ie (r = 2) do
			ie (fld[F_FLGS] & FF_TRIM)
				fld[F_FLGS] := fld[F_FLGS] & ~FF_TRIM;
			else
				fld[F_FLGS] := fld[F_FLGS] | FF_TRIM;
		end
		else if (r = 3) do
			leave;
		end
	end
	drawlines(y, y+6);
	return r >= 0;
end

makefield(x, y) do var new;
	new := newfield(x, y, 1);
	if (\new) return;
	if (	chg_field_type(new) = 0 \/
		chg_field_name(new) = 0 \/
		new[F_TYPE] = T_REF /\ chg_field_length(new) = 0 \/
		new[F_TYPE] = T_REF /\ chg_field_flags(new) = 0 \/
		new[F_TYPE] \= T_REF /\ new[F_LEN] = 0
	) do
		deletefield(new, 1);
		return 0;
	end
	return new;
end

editfield(fld) do var r;
	if (chg_field_name(fld) = 0) return;
	if (fld[F_TYPE] = T_REF /\ chg_field_length(fld) = 0) return;
	if (fld[F_TYPE] = T_REF) chg_field_flags(fld);
end

findfield(x, y) do var fld;
	fld := Sheet[y];
	while (fld) do
		if (fld[F_X] <= x /\ fld[F_X] + fld[F_LEN] > x)
			return fld;
		fld := fld[F_NEXT];
	end
	return 0;
end

fieldmenu(x, y) do var fld, r, i;
	fld := findfield(x, y);
	if (\fld) return makefield(x, y);
	x := fld[F_X];
	y := fld[F_Y]-Top;
	if (x < 0) x := 0;
	if (x + 21 >= Width) x := Width-21;
	if (x + 25 > Xmax) x := Xmax-25;
	if (y < 0) y := 0;
	if (y >= Mtop-8) y := Mtop-9;
	hintbar("Manipulate field at cursor position");
	r := mnu.select(x+Indent, y+Topline, 0,
	[ (C_menutext), (C_menutext), (C_menutext), (C_menutext),
	  (C_menubar) ],
	O_borderc, [
		[ 'e', "Edit field    [E]" ],
		[ 'c', "Copy field    [C]" ],
		[ 'm', "Move field    [M]" ],
		[ 'd', "Delete field  [D]" ],
	0 ]);
	drawlines(y, y+6);
	ie (r = 0) editfield(fld);
	else ie (r = 1) Floating := copyfield(fld);
	else ie (r = 2) Floating := unlinkfield(fld);
	else if (r = 3) deletefield(fld, 1);
end

inside(pt, fld) return pt >= fld[F_X] /\ pt < fld[F_X]+fld[F_LEN];

overlap(line, fld) do
	while (line) do
		if (	inside(fld[F_X], line) \/
			inside(fld[F_X]+fld[F_LEN]-1, line) \/
			inside(line[F_X], fld)
		)
			return 1;
		line := line[F_NEXT];
	end
	return 0;
end

unfloat() do
	if (overlap(Sheet[Ypos], Floating)) do
		message("Overlapping fields");
		return;
	end
	linkfield(Ypos, Floating);
	drawline(Ypos-Top, Sheet[Ypos]);
	Floating := 0;
end

ycenter(n) return (Mtop-3)/2 - n/2;

confirm(yes, no, hint) do var r, i;
	hintbar(hint);
	r := mnu.select(Xmax/2-8, ycenter(4)+Topline, 0,
	[ (C_menutext), (C_menutext), (C_menutext), (C_menutext),
	  (C_menubar) ],
	O_borderc, [
		[ 'y', (yes) ],
		[ 'n', (no)  ],
	0 ]);
	drawlines(ycenter(4), ycenter(4)+4);
	return r = 0;
end

abandon() do
	if (\Modified) return 1;
	return confirm(
		"[Y] lose changes",
		"[N] keep changes",
		"About to abandon changes to this sheet");
end

registersheet(name) do var k;
	k := str.length(name)+1;
	if (Snt + k >= (NAMELEN+1)*MAXSHEETS) do
		message("cannot register sheet (ouf of name space)");
		return;
	end
	str.copy(@Shndxbuf::Snt, name);
	Sheetindex[Nsheets] := @Shndxbuf::Snt;
	Snt := Snt+k;
	Sheetnlist[Nsheets] := @Snlbuf[NSheets*2];
	Sheetnlist[Nsheets][0] := 'i'+'0';
	Sheetnlist[Nsheets][1] := Sheetindex[Nsheets];
	Nsheets := Nsheets+1;
	Sheetnlist[Nsheets] := 0;
end

savesheet(nosave) do
	object	shs[iostream], tmp[iostream];
	var	shbuf::BUFSIZE, tmpbuf::BUFSIZE;
	var	tmpname::PATHLEN+2;
	var	fbuf::NCOLS;
	var	fld;
	var	k, i;
	var	pass;
	var	uflag;

	uflag := nosave;
	str.copy(tmpname, Sheetpath);
	k := str.length(tmpname);
	str.copy(@tmpname::k, "-");
	if (t.close(t.open(Sheetpath, T3X.OREAD)) < 0)
		t.close(t.open(Sheetpath, T3X.OWRITE));
	t.remove(tmpname);
	t.rename(Sheetpath, tmpname);
	if (shs.open(Sheetpath, shbuf, BUFSIZE, iostream.FWRITE) = %1) do
		t.rename(tmpname, Sheetpath);
		message("failed to create sheet file");
		return -1;
	end
	if (\nosave) do
		shs.wrch('A'-'@');
		shs.writes(Sheetname);
		shs.writes("\n");
		str.format(fbuf, "%D %D %D %D %D\n", [
			(Indent), (Width), (Spacing), (Pagesize), (Absmode) ]);
		shs.writes(fbuf);
		for (i=0, MAXLINES) do
			fld := Sheet[i];
			while (fld) do
				str.format(fbuf, "%D %D %D %D %D [%S]\n",
				[ (fld[F_X]), (fld[F_Y]), (fld[F_TYPE]),
			 	  (fld[F_LEN]), (fld[F_FLGS]), (fld[F_NAME])
				]);
				shs.writes(fbuf);
				fld := fld[F_NEXT];
			end
		end
	end
	if (tmp.open(tmpname, tmpbuf, BUFSIZE, iostream.FREAD) = %1) do
		shs.close();
		t.remove(Sheetpath);
		t.rename(tmpname, Sheetpath);
		message("OOPS! Failed to reopen workfile (could not save)");
		return -1;
	end
	nosave := nosave-> nosave: Sheetname;
	pass := 1;
	while (1) do
		k := tmp.reads(fbuf, NCOLS);
		if (tmp.eof()) leave;
		if (fbuf::0 = 'A'-'@') do
			fbuf::(k-1) := 0;
			if (k > 1 /\ fbuf::(k-2) = '\r') do
				k := k-1;
				fbuf::(k-1) := 0;
			end
			pass := str.comp(@fbuf::1, nosave);
			fbuf::(k-1) := '\n';
		end
		if (pass) shs.writes(fbuf);
	end
	tmp.close();
	shs.close();
	!XXX remove(tmpname); ! keep the workfile for now
	if (\uflag) Modified := 0;
end

removesheet() do var r, i;
	if (\Nsheets) do
		message("No sheets defined");
		return;
	end
	hintbar("Select sheet to remove");
	r := mnu.select(Indent, Topline, 0,
	[ (C_menutext), (C_menutext), (C_menutext), (C_menutext),
  	(C_menubar) ],
	O_borderc, Sheetnlist);
	if (r < 0) return;
	if (\str.comp(Sheetname, Sheetindex[r])) do
		message("Cannot delete sheet in editor");
		return;
	end
	savesheet(Sheetindex[r]);
	for (i=r, Nsheets) do
		Sheetnlist[i] := Sheetnlist[i+1];
		Sheetindex[i] := Sheetindex[i+1];
	end
	Nsheets := Nsheets-1;
end

erasesheet() do
	if (\confirm(
		"[Y] erase sheet",
		"[N] keep sheet ",
		"About to erase all fields on the sheet")
	)
		return;
	initsheet();
	! Modification flag intentionally not set.
end

newsheet() do var k, i, buf::NAMELEN+1;
	if (Nsheets >= MAXSHEETS) do
		message("Too many sheets for this database");
		return;
	end
	if (\abandon()) return;
	buf::0 := 0;
	while (1) do
		hintbar("Enter a name for the new sheet");
		k := prm.input(Indent+4, Topline-1, NAMELEN, NAMELEN+1,
			[ (C_browsetext), (C_browsetext), (C_browsetext),
		  	(C_browsetext), (C_browsebar) ],
			buf, [ 2, '\r', ttyctl.K_ESC ]);
		if (k \= '\r' \/ \buf::0) return;
		for (i=0, Nsheets)
			if (\str.comp(buf, Sheetindex[i])) do
				message(
				"Name already in use -- press [ENTER]");
				tty.readc();
				message(0);
				leave;
			end
		if (i < Nsheets) loop;
		leave;
	end
	if (\confirm(
		"[Y] Copy fields   ",
		"[N] Clear sheet   ",
		"Start with an empty sheet or copy fields to new sheet")
	)
		initsheet();
	registersheet(buf);
	str.copy(Sheetname, buf);
	savesheet(0);
end

loadsheet(s) do
	var	r;
	object	shs[iostream];
	var	shbuf::BUFSIZE;
	var	name::NAMELEN+1;
	var	fbuf::NCOLS;
	var	k, ok;
	var	new;

	if (Nsheets < 1) do
		if (\s) message("No sheets defined for this database");
		return;
	end
	ie (s) do
		for (r=0, Nsheets)
			if (\str.comp(s, Sheetindex[r]))
				leave;
		if (r >= Nsheets) return;
	end
	else do
		if (\abandon()) return;
		hintbar("Select sheet to load from the menu");
		r := mnu.select(Indent+3, Topline, 0,
		[ (C_menutext), (C_menutext), (C_menutext), (C_menutext),
	  	(C_menubar) ],
		O_borderc, Sheetnlist);
		if (r < 0) return;
	end
	initsheet();
	if (shs.open(Sheetpath, shbuf, BUFSIZE, iostream.FREAD) = %1) do
		message("OOPS! Cannot open sheet file");
		return 0;
	end
	while (1) do
		k := shs.reads(fbuf, NCOLS);
		if (shs.eof()) leave;
		if (fbuf::0 = 'A'-'@') do
			fbuf::(k-1) := 0;
			if (k > 1 /\ fbuf::(k-2) = '\r') fbuf::(k-2) := 0;
			if (\str.comp(@fbuf::1, Sheetindex[r])) leave;
		end
	end
	if (shs.eof()) do
		message("OOPS! Sheet not available");
		return 0;
	end
	str.copy(Sheetname, Sheetindex[r]);
	k := shs.reads(fbuf, NCOLS);
	ok := str.parse(fbuf, "%D %D %D %D %D", [
		@Indent, @Width, @Spacing, @Pagesize, @Absmode ]) = 5;
	if (ok) while (1) do
		ok := 1;
		k := shs.reads(fbuf, NCOLS);
		if (shs.eof() \/ fbuf::0 = 'A'-'@') leave;
		new := newfield(0, 0, 0);
		ok := 0;
		if (str.parse(fbuf, "%D %D %D %D %D [%S", [
			(@new[F_X]), (@new[F_Y]), (@new[F_TYPE]),
			(@new[F_LEN]), (@new[F_FLGS]), (name)
		]) \= 6) leave;
		k := str.rscan(name, ']');
		if (k < 0) leave;
		name::k := 0;
		ie (new[F_TYPE] = T_LIT)
			new[F_NAME] := savestr(name);
		else
			new[F_NAME] := findref(name);
		linkfield(new[F_Y], new);
	end
	shs.close();
	if (\ok) do
		message("Corrupt sheet (cleared)");
		initsheet();
	end
	Modified := 0;
	if (\s) do
		clear();
		drawsheet();
	end
end

checkwidth(n) do
	!XXX
end

setupsheet() do
	var	setupform, boolmenu;
	var	bwidth::30, bindent::30, bspacing::30, bpagesize::30;
	var	nwidth, nindent, nspacing, npagesize;
	var	failed;

	hintbar("Edit sheet settings: [Up/Down]=select, [ENTER]=change");
	str.numtostr(bindent, Indent, 10);
	str.numtostr(bwidth, Width, 10);
	str.numtostr(bspacing, Spacing, 10);
	str.numtostr(bpagesize, Pagesize, 10);
	boolmenu := [
		[ 'd', "Yes" ],
		[ 'e', "No " ],
	0 ];
	setupform := [
	[ form.T_TEXT,	 5, 1, 0,	"Sheet Setup"		],
	[ form.T_TEXT,	 5, 3, 0,	"Indent (columns)"	],
	[ form.T_INPUT,	28, 3, (@bindent), [3,3]		],
	[ form.T_TEXT,	 5, 4, 0,	"Width (columns)"	],
	[ form.T_INPUT,	28, 4, (@bwidth), [3,3]			],
	[ form.T_TEXT,	 5, 5, 0,	"Spacing (lines)"	],
	[ form.T_INPUT,	28, 5, (@bspacing), [3,3]		],
	[ form.T_TEXT,	 5, 6, 0,	"Page size (lines)"	],
	[ form.T_INPUT,	28, 6, (@bpagesize), [3,3]		],
	[ form.T_TEXT,	 5, 7, 0,	"Repeat records"	],
	[ form.T_MENU,	28, 7, (@Absmode),(boolmenu)		],
	[ form.T_DONE,	15, 9, 0,	"[Done]"		],
	0 ];
	frm.complete(Xmax/2-18, ycenter(11)+Topline, 36, 11, 
	[ (C_menutext), (C_menutext), (C_menutext), (C_menutext),
	  (C_menubar) ],
	[ (C_edittext), (C_edittext), (C_edittext), (C_edittext),
	  (C_editbar) ],
	O_borderc, 
	[ ttyctl.K_DOWN, ttyctl.K_UP, ttyctl.K_ESC ], setupform);
	Modified := 1;
	drawlines(ycenter(11), ycenter(11)+11);
	nwidth := str.strtonum(bwidth, 10, 0);
	nindent := str.strtonum(bindent, 10, 0);
	nspacing := str.strtonum(bspacing, 10, 0);
	npagesize := str.strtonum(bpagesize, 10, 0);
	failed := 0;
	ie (checkwidth(nwidth))
		failed := 1;
	else
		Width := nwidth;
	ie (nindent > 0 /\ nindent + Width > Xmax)
		failed := 1;
	else
		Indent := nindent;
	ie (nspacing < 0 \/ nspacing > 60)
		failed := 0;
	else
		Spacing := nspacing;
	ie (npagesize < 0)
		failed := 0;
	else
		Pagesize := npagesize;
	clear();
	if (failed) message("Warning: out-of-range values ignored");
end

position(recno) do var kmask, hpos, lpos;
	kmask := 16384;
	kmask := (kmask << 2) - 1;
	hpos := recno >> 6;
	lpos := recno << 10 & kmask;
	t.seek(Dbfd, 0, T3X.SEEK_SET);
	while (hpos) do
		t.seek(Dbfd, 16384, T3X.SEEK_FWD);
		t.seek(Dbfd, 16384, T3X.SEEK_FWD);
		t.seek(Dbfd, 16384, T3X.SEEK_FWD);
		t.seek(Dbfd, 16384, T3X.SEEK_FWD);
		hpos := hpos-1;
	end
	t.seek(Dbfd, lpos & kmask, T3X.SEEK_FWD);
end

scatter(recno) do var i, j, n;
	position(recno);
	if (t.read(Dbfd, Record, RECORDSIZE) \= RECORDSIZE) do
		message("Database read error");
		return -1;
	end
	j := 0;
	n := 0;
	for (i=0, RECORDSIZE) do
		if (Record::i = '\n') do
			Dataset[n] := @Record::j;
			n := n+1;
			Record::i := 0;
			j := i+1;
		end
		if (n >= Nfields) leave;
	end
end

lookupfield(s) do var i;
	for (i=0, Nfields)
		if (\str.comp(s, Taglist[i]))
			return i;
	return -1;
end

writefield(fld) do var text, k, fno;
	ie (fld[F_TYPE] = T_REF) do
		fno := lookupfield(fld[F_NAME]);
		if (fno < 0) do
			message("unknown field name in sheet");
			return -1;
		end
		text := Dataset[fno];
	end
	else do
		text := fld[F_NAME];
	end
	k := str.length(text);
	ie (fld[F_FLGS] & FF_TRIM /\ k >= fld[F_LEN]) do
		Rpt.write(text, fld[F_LEN]);
		return fld[F_LEN];
	end
	else do
		Rpt.writes(text);
		if (fld[F_FLGS] & FF_FILL)
			Rpt.writes(@Blanks::(NCOLS-fld[F_LEN]+k));
		return fld[F_LEN];
	end
end

render() do
	var 	i, j, k, first, last;
	var	fields[NCOLS], nf, p;
	var	col;

	first := MAXLINES;
	last := 0;
	ie (Absmode) do
		first := 0;
		last := Pagesize-1;
		if (last >= MAXLINES) last := MAXLINES-1;
	end
	else for (i=0, MAXLINES) do
		if (Sheet[i]) do
			if (i < first) first := i;
			if (i > last) last := i;
		end
	end
	if (\Absmode) do
		Row := Row+Spacing+last-first+1;
		if (Row >= Pagesize) do
			Row := 0;
			Rpt.wrch('\f');
		end
	end
	for (i=0, Spacing)
		Rpt.writes(O_rg_addcr-> "\r\n": "\n");
	for (i=first, last+1) do
		p := Sheet[i];
		nf := 0;
		while (p) do
			for (j=0, nf)
				if (fields[j][F_X] > p[F_X])
					leave;
			t.memcopy(@fields[j+1], @fields[j], (nf-j)*t.bpw());
			fields[j] := p;
			nf := nf+1;
			p := p[F_NEXT];
		end
		col := 0;
		if (Indent) Rpt.writes(@Blanks::(NCOLS-Indent));
		for (j=0, nf) do
			Rpt.writes(@Blanks::(NCOLS-fields[j][F_X]+col));
			k := writefield(fields[j]);
			if (k < 0) return -1;
			col := fields[j][F_X] + k;
		end
		Rpt.writes(O_rg_addcr-> "\r\n": "\n");
	end
	if (Absmode) Rpt.wrch('\f');
end

print(recno) do
	if (scatter(recno) < 0) return -1;
	if (render() < 0) return -1;
end

spawnprog(prog) do var r;
	tty.color(ttyctl.F_GREY|ttyctl.B_BLACK);
	tty.clear();
	tty.mode(0);
	if (O_dosflag) str.xlate(Rptfile, '/', '\\');
	r := sys.spawn(prog, [ (Rptfile), 0 ], system.SPAWN_WAIT);
	tty.mode(1);
	clear();
	if (r) message("Command failed -- check config");
end

savereport() do var k, buffer::PATHLEN+5;
	str.copy(buffer, Rptfile);
	message(0);
	tty.move(Indent, Ymax-4);
	tty.color(C_background);
	tty.writes("File:");
	k := prm.input(Indent+6, Mtop, Width-6, PATHLEN,
		[ (C_menutext), (C_menutext), (C_menutext), (C_menutext),
		  (C_menubar) ],
		buffer, [ 2, '\r', ttyctl.K_ESC ]);
	if (k = ttyctl.K_ESC \/ \str.comp(buffer, Rptfile)) return;
	k := str.scan(buffer, '.');
	if (k >= 0) buffer::k := 0;
	k := str.length(buffer);
	str.copy(@buffer::k, ".rpt");
	t.remove(buffer);
	if (t.rename(Rptfile, buffer) < 0)
		message("Failed to save file");
end

andnow() do var r;
	hintbar("Report complete. And now?");
	r := mnu.select(Xmax/2-9, ycenter(4)+Topline, 0,
	[ (C_menutext), (C_menutext), (C_menutext), (C_menutext),
	  (C_menubar) ],
	O_borderc, [
		[ 'v', "View           [V]" ],
		[ 's', "Save to file   [S]" ],
		[ 'p', "Print          [P]" ],
		[ 'f', "Forget it      [F]" ],
	0 ]);
	drawlines(ycenter(4), ycenter(4)+6);
	ie (r = 0) spawnprog(O_rg_view);
	else ie (r = 1) savereport();
	else ie (r = 2) spawnprog(O_rg_print);
	else message("Report kept in .rpt file");
end

generate() do
	var	k, i;
	var	ndxbuf[NDXLEN];
	var	tmp::NCOLS+1;
	var	err;

	str.format(Rptfile, "%S.rpt", [(Basename)]);
	if (Rpt.open(Rptfile, Rptbuf, BUFSIZE, iostream.FWRITE) = %1) do
		message("Cannot open report file");
		return -1;
	end
	i := 1;
	t.seek(Ndxfd, 0, T3X.SEEK_SET);
	t.seek(Dbfd, 0, T3X.SEEK_SET);
	hintbar("Generating");
	err := 0;
	Row := 0;
	while (1) do
		k := t.read(Ndxfd, ndxbuf, NDXLEN*t.bpw());
		if (k = 0) leave;
		if (k \= NDXLEN * t.bpw()) do
			message("Index read error");
			return;
		end
		if (Tagtbl::i) do
			str.format(tmp, "Generating record #%D", [(i)]);
			_hintbar(tmp);
			err := print(ndxbuf[I_KEY]);
			if (err < 0) leave;
		end
		i := i+1;
	end
	if (Row) Rpt.wrch('\f');
	Rpt.close();
	if (\err) andnow();
end

mainmenu() do var r, i;
	hintbar("Main menu: manage sheets");
	r := mnu.select(Xmax/2-10, ycenter(10)+Topline, 0,
	[ (C_menutext), (C_menutext), (C_menutext), (C_menutext),
	  (C_menubar) ],
	O_borderc, [
		[ 'l', "Load sheet      [L]" ],
		[ 's', "Save sheet      [S]" ],
		[ 'p', "Print report    [P]" ],
		[ 'u', "Setup sheet     [U]" ],
		[ 'e', "Erase fields    [E]" ],
		[ 'n', "New sheet       [N]" ],
		[ 'r', "Remove sheet    [R]" ],
		[ 'q', "Quit to Node    [Q]" ],
	0 ]);
	drawlines(ycenter(10), ycenter(10)+12);
	ie (r = 0) loadsheet(0);
	else ie (r = 1) savesheet(0);
	else ie (r = 2) generate();
	else ie (r = 3) setupsheet();
	else ie (r = 4) erasesheet();
	else ie (r = 5) newsheet();
	else ie (r = 6) removesheet();
	else if (r = 7) return 1;
	return 0;
end

showpos() do var buf::NCOLS+1;
	str.format(buf, "C:%D L:%D T:%D O:%D%S", [(Xpos+1), (Ypos+1),
		(Top), (Offset), (Modified-> " *": "")]);
	tty.move(Indent, Mtop);
	tty.color(C_background);
	tty.clreol();
	tty.writes(buf);
	Posted := 0;
end

editsheet() do
	var	red;
	var	k;
	var	pgsize;
	var	hint, floathint;
	var	fld;

	hint := "[Arrow keys]=navigate, [ENTER]=edit, [ESC]=menu";
	floathint := "Floating: [Arrow keys]=navigate, [ENTER]=place";
	red := -1;
	while (1) do
		if (Xpos < Offset) do
			Offset := 0;
			red := 2;
		end
		if (Xpos-Offset >= Xmax-Indent) do
			Offset := Width-Xmax+Indent;
			red := 2;
		end
		if (Floating /\ Xpos + Floating[F_LEN] > Xmax-Indent) do
			Offset := Width-Xmax+Indent;
			red := 2;
		end
		if (red) do
			pgsize := Mtop-4;
			if (red < 0) showframe();
			if (red & 1) hintbar(Floating-> floathint: hint);
			if (red & 2) drawsheet();
			red := 0;
		end
		if (\Posted) showpos();
		if (Floating) do
			Floating[F_X] := Xpos;
			Floating[F_Y] := Ypos;
			drawfield(Floating);
		end
		tty.move(Xpos+Indent-Offset, Ypos-Top+Topline);
		k := tty.readc();
		if (Posted) showpos();
		if (Floating) drawline(Ypos-Top, Sheet[Ypos]);
		ie (k = ttyctl.K_RIGHT \/ k = 'F'-'@') do
			Xpos := Xpos+1;
			if (Floating /\ Xpos + Floating[F_LEN] > Width)
				Xpos := 0;
			if (Xpos >= Width) Xpos := 0;
		end
		else ie (k = ttyctl.K_LEFT \/ k = 'B'-'@') do
			Xpos := Xpos-1;
			if (Xpos < 0)
				ie (Floating)
					Xpos := Width-Floating[F_LEN];
				else
					Xpos := Width-1;
		end
		else ie (k = ttyctl.K_HOME \/ k = 'A'-'@') do
			Xpos := 0;
		end
		else ie (k = ttyctl.K_DOWN \/ k = 'N'-'@') do
			Ypos := Ypos+1;
			if (Ypos >= MAXLINES) Ypos := Ypos-1;
			if (Ypos - Top >= Mtop-3) do
				Top := Top+1;
				red := 3;
			end
		end
		else ie (k = ttyctl.K_UP \/ k = 'P'-'@') do
			Ypos := Ypos-1;
			if (Ypos < 0) Ypos := 0;
			if (Ypos < Top) do
				Top := Top-1;
				red := 3;
			end
		end
		else ie (k = ttyctl.K_NEXT \/ k = 'V'-'@') do
			Ypos := Ypos + pgsize;
			if (Ypos >= MAXLINES) Ypos := MAXLINES-1;
			Top := Top + pgsize;
			if (Top >= MAXLINES-pgsize) Top := MAXLINES-pgsize-1;
			red := 3;
		end
		else ie (k = ttyctl.K_PREV \/ k = 'T'-'@') do
			Ypos := Ypos - pgsize;
			if (Ypos < 0) Ypos := 0;
			Top := Top - pgsize;
			if (Top < 0) Top := 0;
			red := 3;
		end
		else ie (k = '\r') do
			if (Floating) do
				unfloat();
				red := 1;
				loop;
			end
			fieldmenu(Xpos, Ypos);
			Modified := 1;
			red := 3;
		end
		else ie (k = 'L'-'@') do
			red := -1;
		end
		else ie (k = ttyctl.K_ESC \/ k = ':') do
 			red := 3;
			if (Floating) do
				deletefield(Floating, 0);
				Floating := 0;
				loop;
			end
			if (mainmenu() /\ abandon()) leave;
		end
		else do
			if (Floating) loop;
			fld := findfield(Xpos, Ypos);
			if (\fld) loop;
			ie (k = 'D'-'@') do
				deletefield(fld, 1);
				drawline(Ypos-Top, Sheet[Ypos]);
			end
			else ie (k = 'c') Floating := copyfield(fld);
			else if (k = 'm') Floating := unlinkfield(fld);
		end
	end
end

error(msg) do var b::10;
	fini();
	t.write(T3X.SYSERR, "NODE_RG: ", 9);
	t.write(T3X.SYSERR, msg, str.length(msg));
	t.write(T3X.SYSERR, t.newline(b), str.length(b));
	halt 1;
end

loadindex() do
	object	shs[iostream];
	var	shbuf::BUFSIZE;
	var	fbuf::NCOLS;
	var	k, i;

	Nsheets := 0;
	Snt := 0;
	if (shs.open(Sheetpath, shbuf, BUFSIZE, iostream.FREAD) = %1)
		return;
	while (1) do
		k := shs.reads(fbuf, NCOLS);
		if (shs.eof()) leave;
		if (fbuf::0 = 'A'-'@') do
			fbuf::(k-1) := 0;
			if (k > 1 /\ fbuf::(k-2) = '\r') fbuf::(k-2) := 0;
			registersheet(@fbuf::1);
		end
	end
	shs.close();
end

load() do var ln, k, i, n, tmppath::PATHLEN+1;
	if (	t.getarg(1, Basename, PATHLEN+1) < 0 \/
		t.getarg(2, Confpath, PATHLEN+1) < 0 \/
		t.getarg(3, Record, 64) >= 0
	)
		error("Usage: node_rg database config-file");
	if (Cnf.load(Confpath, @ln) \= 0)
		error("error in config file");
	str.format(Dbpath, "%S.db", [(Basename)]);
	str.format(Sheetpath, "%S.rs", [(Basename)]);
	Dbfd := t.open(Dbpath, T3X.OREAD);
	if (Dbfd < 0) error("could not open database");
	if (t.read(Dbfd, Descriptor, RECORDSIZE) \= RECORDSIZE)
		error("corrupt database");
	Nfields := (Descriptor::6 - '0') * 10 + Descriptor::7 - '0';
	for (i=15, RECORDSIZE)
		if (Descriptor::i = '\n') leave;
	i := i+1;
	if (i >= RECORDSIZE) error("corrupt database");
	k := i;
	n := 0;
	for (i=i, RECORDSIZE) do
		if (Descriptor::i = '\n') do
			Descriptor::i := 0;
			Taglist[n] := @Descriptor::(k+1);
			n := n+1;
			k := i+1;
		end
		if (n >= Nfields) leave;
	end
	for (i=0, Nfields) do
		Fieldnames[i] := @Fmbuffer[i*2];
		Fieldnames[i][0] := 'i'+'0';
		Fieldnames[i][1] := Taglist[i];
	end
	Fieldnames[i] := 0;
	str.format(tmppath, "%S.x", [(Basename)]);
	Ndxfd := t.open(tmppath, T3X.OREAD);
	if (Ndxfd < 0) error("missing index");
	str.format(tmppath, "%S.tg", [(Basename)]);
	Tagfd := t.open(tmppath, T3X.OREAD);
	if (Tagfd < 0) error("missing tag file");
	if (t.read(Tagfd, Tagtbl, TAGTBLSIZE-1) \= TAGTBLSIZE-1)
		error("invalid tag file");
	t.close(Tagfd);
	loadindex();
end

do
	init();
	load();
	setup();
	ie (\Nsheets)
		initsheet();
	else
		loadsheet("*default*");
	editsheet();
	fini();
end
